/*
 * Library to access the Windows Explorer thumbnail cache database (WTCDB) format
 *
 * Copyright (C) 2010-2024, Joachim Metz <joachim.metz@gmail.com>
 *
 * Refer to AUTHORS for acknowledgements.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#if !defined( _LIBWTCDB_H )
#define _LIBWTCDB_H

#include <libwtcdb/codepage.h>
#include <libwtcdb/definitions.h>
#include <libwtcdb/error.h>
#include <libwtcdb/extern.h>
#include <libwtcdb/features.h>
#include <libwtcdb/types.h>

#include <stdio.h>

#if defined( LIBWTCDB_HAVE_BFIO )
#include <libbfio.h>
#endif

#if defined( __cplusplus )
extern "C" {
#endif

/* -------------------------------------------------------------------------
 * Support functions
 * ------------------------------------------------------------------------- */

/* Returns the library version
 */
LIBWTCDB_EXTERN \
const char *libwtcdb_get_version(
             void );

/* Returns the access flags for reading
 */
LIBWTCDB_EXTERN \
int libwtcdb_get_access_flags_read(
     void );

/* Retrieves the narrow system string codepage
 * A value of 0 represents no codepage, UTF-8 encoding is used instead
 * Returns 1 if successful or -1 on error
 */
LIBWTCDB_EXTERN \
int libwtcdb_get_codepage(
     int *codepage,
     libwtcdb_error_t **error );

/* Sets the narrow system string codepage
 * A value of 0 represents no codepage, UTF-8 encoding is used instead
 * Returns 1 if successful or -1 on error
 */
LIBWTCDB_EXTERN \
int libwtcdb_set_codepage(
     int codepage,
     libwtcdb_error_t **error );

/* Determines if a file contains a WTCDB file signature
 * Returns 1 if true, 0 if not or -1 on error
 */
LIBWTCDB_EXTERN \
int libwtcdb_check_file_signature(
     const char *filename,
     libwtcdb_error_t **error );

#if defined( LIBWTCDB_HAVE_WIDE_CHARACTER_TYPE )

/* Determines if a file contains a WTCDB file signature
 * Returns 1 if true, 0 if not or -1 on error
 */
LIBWTCDB_EXTERN \
int libwtcdb_check_file_signature_wide(
     const wchar_t *filename,
     libwtcdb_error_t **error );

#endif /* defined( LIBWTCDB_HAVE_WIDE_CHARACTER_TYPE ) */

#if defined( LIBWTCDB_HAVE_BFIO )

/* Determines if a file contains a WTCDB file signature using a Basic File IO (bfio) handle
 * Returns 1 if true, 0 if not or -1 on error
 */
LIBWTCDB_EXTERN \
int libwtcdb_check_file_signature_file_io_handle(
     libbfio_handle_t *file_io_handle,
     libwtcdb_error_t **error );

#endif /* defined( LIBWTCDB_HAVE_BFIO ) */

/* -------------------------------------------------------------------------
 * Notify functions
 * ------------------------------------------------------------------------- */

/* Sets the verbose notification
 */
LIBWTCDB_EXTERN \
void libwtcdb_notify_set_verbose(
      int verbose );

/* Sets the notification stream
 * Returns 1 if successful or -1 on error
 */
LIBWTCDB_EXTERN \
int libwtcdb_notify_set_stream(
     FILE *stream,
     libwtcdb_error_t **error );

/* Opens the notification stream using a filename
 * The stream is opened in append mode
 * Returns 1 if successful or -1 on error
 */
LIBWTCDB_EXTERN \
int libwtcdb_notify_stream_open(
     const char *filename,
     libwtcdb_error_t **error );

/* Closes the notification stream if opened using a filename
 * Returns 0 if successful or -1 on error
 */
LIBWTCDB_EXTERN \
int libwtcdb_notify_stream_close(
     libwtcdb_error_t **error );

/* -------------------------------------------------------------------------
 * Error functions
 * ------------------------------------------------------------------------- */

/* Frees an error
 */
LIBWTCDB_EXTERN \
void libwtcdb_error_free(
      libwtcdb_error_t **error );

/* Prints a descriptive string of the error to the stream
 * Returns the number of printed characters if successful or -1 on error
 */
LIBWTCDB_EXTERN \
int libwtcdb_error_fprint(
     libwtcdb_error_t *error,
     FILE *stream );

/* Prints a descriptive string of the error to the string
 * Returns the number of printed characters if successful or -1 on error
 */
LIBWTCDB_EXTERN \
int libwtcdb_error_sprint(
     libwtcdb_error_t *error,
     char *string,
     size_t size );

/* Prints a backtrace of the error to the stream
 * Returns the number of printed characters if successful or -1 on error
 */
LIBWTCDB_EXTERN \
int libwtcdb_error_backtrace_fprint(
     libwtcdb_error_t *error,
     FILE *stream );

/* Prints a backtrace of the error to the string
 * Returns the number of printed characters if successful or -1 on error
 */
LIBWTCDB_EXTERN \
int libwtcdb_error_backtrace_sprint(
     libwtcdb_error_t *error,
     char *string,
     size_t size );

/* -------------------------------------------------------------------------
 * File functions
 * ------------------------------------------------------------------------- */

/* Creates a file
 * Make sure the value file is referencing, is set to NULL
 * Returns 1 if successful or -1 on error
 */
LIBWTCDB_EXTERN \
int libwtcdb_file_initialize(
     libwtcdb_file_t **file,
     libwtcdb_error_t **error );

/* Frees a file
 * Returns 1 if successful or -1 on error
 */
LIBWTCDB_EXTERN \
int libwtcdb_file_free(
     libwtcdb_file_t **file,
     libwtcdb_error_t **error );

/* Signals the libwtcdb file to abort its current activity
 * Returns 1 if successful or -1 on error
 */
LIBWTCDB_EXTERN \
int libwtcdb_file_signal_abort(
     libwtcdb_file_t *file,
     libwtcdb_error_t **error );

/* Opens a file
 * Returns 1 if successful or -1 on error
 */
LIBWTCDB_EXTERN \
int libwtcdb_file_open(
     libwtcdb_file_t *file,
     const char *filename,
     int access_flags,
     libwtcdb_error_t **error );

#if defined( LIBWTCDB_HAVE_WIDE_CHARACTER_TYPE )

/* Opens a file
 * Returns 1 if successful or -1 on error
 */
LIBWTCDB_EXTERN \
int libwtcdb_file_open_wide(
     libwtcdb_file_t *file,
     const wchar_t *filename,
     int access_flags,
     libwtcdb_error_t **error );

#endif /* defined( LIBWTCDB_HAVE_WIDE_CHARACTER_TYPE ) */

#if defined( LIBWTCDB_HAVE_BFIO )

/* Opens a file using a Basic File IO (bfio) handle
 * Returns 1 if successful or -1 on error
 */
LIBWTCDB_EXTERN \
int libwtcdb_file_open_file_io_handle(
     libwtcdb_file_t *file,
     libbfio_handle_t *file_io_handle,
     int access_flags,
     libwtcdb_error_t **error );

#endif /* defined( LIBWTCDB_HAVE_BFIO ) */

/* Closes a file
 * Returns 0 if successful or -1 on error
 */
LIBWTCDB_EXTERN \
int libwtcdb_file_close(
     libwtcdb_file_t *file,
     libwtcdb_error_t **error );

/* Retrieves the file type
 * Returns 1 if successful or -1 on error
 */
LIBWTCDB_EXTERN \
int libwtcdb_file_get_type(
     libwtcdb_file_t *file,
     uint8_t *type,
     libwtcdb_error_t **error );

/* Retrieves the number of items
 * Returns 1 if successful or -1 on error
 */
LIBWTCDB_EXTERN \
int libwtcdb_file_get_number_of_items(
     libwtcdb_file_t *file,
     int *number_of_items,
     libwtcdb_error_t **error );

/* Retrieves a specific item
 * Returns 1 if successful or -1 on error
 */
LIBWTCDB_EXTERN \
int libwtcdb_file_get_item(
     libwtcdb_file_t *file,
     int item_index,
     libwtcdb_item_t **item,
     libwtcdb_error_t **error );

/* -------------------------------------------------------------------------
 * Item functions
 * ------------------------------------------------------------------------- */

/* Frees an item
 * Returns 1 if successful or -1 on error
 */
LIBWTCDB_EXTERN \
int libwtcdb_item_free(
     libwtcdb_item_t **item,
     libwtcdb_error_t **error );

#if defined( __cplusplus )
}
#endif

#endif /* !defined( _LIBWTCDB_H ) */

